package rwi.eclipse.team.ccvs.ssl;

/*******************************************************************************
 * Copyright (c) 2003 Rolf Wilms (rolf_wilms@hotmail.com).
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors:
 * Extensions/modifications to original org.eclipse.team.internal.ccvs.core.connection.PServerConnection
 * (c) Copyright Rolf Wilms (rolf_wilms@hotmail.com) 2003. 
 * All Rights Reserved.
 * 
 * This class is based on org.eclipse.team.internal.ccvs.core.connection.PServerConnection which is
 * (c) Copyright IBM Corp. 2000, 2001.
 * All Rights Reserved.
 * 
 * CVS sserver support contributions by Dale King
 * 
 *******************************************************************************/


import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.net.Socket;
import java.net.SocketException;

import javax.net.ssl.SSLSocket;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.osgi.util.NLS;
import org.eclipse.team.internal.ccvs.core.CVSMessages;
import org.eclipse.team.internal.ccvs.core.ICVSRepositoryLocation;
import org.eclipse.team.internal.ccvs.core.IServerConnection;
import org.eclipse.team.internal.ccvs.core.Policy;
import org.eclipse.team.internal.ccvs.core.connection.CVSAuthenticationException;
import org.eclipse.team.internal.ccvs.core.util.Util;

/**
 * @author Rolf Wilms (rolf_wilms@hotmail.com)
 *
 * A connection used to talk to a cvs sserver.
 * 
 * Roughly speaking sserver is pserver via SSL. 
 * This is how it works with CVSNT:
 * 
 * 1) Client opens TCP/IP port on cvs server
 * 2) Clients requests SSL authentication from server
 * 3) Server responds it can do sserver
 * 4) Client (and server) switch the existing TCP/IP port to SSL. No new port is opened.
 * 5) After switching to SSL, communication continues like with pserver, including the "pserver" authentication
 * 6) Communication uses SSL until TCP/IP connection is closed.    
 * 
 * With CVS sserver it is a little bit different:
 * 
 * 1) Client opens TCP/IP port on cvs server
 * 2) Client (and server) switch the existing TCP/IP port to SSL. No new port is opened.
 * 3) Clients requests SSL authentication from server
 * 4) Communication continues like with pserver, including the "pserver" authentication
 * 5) Communication uses SSL until TCP/IP connection is closed.    
 * 
 */
public class SSLServerConnection implements IServerConnection {

	public static final char NEWLINE = 0xA;

	/** default CVS pserver port */
	private static final int DEFAULT_PORT = 2401;

	/** error line indicators */
	private static final char ERROR_CHAR = 'E';
	private static final String ERROR_MESSAGE = "error 0"; //$NON-NLS-1$
	private static final String NO_SUCH_USER = "no such user"; //$NON-NLS-1$

	private static final char[] SCRAMBLING_TABLE=new char[] {
	0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
	16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
	114,120,53,79,96,109,72,108,70,64,76,67,116,74,68,87,
	111,52,75,119,49,34,82,81,95,65,112,86,118,110,122,105,
	41,57,83,43,46,102,40,89,38,103,45,50,42,123,91,35,
	125,55,54,66,124,126,59,47,92,71,115,78,88,107,106,56,
	36,121,117,104,101,100,69,73,99,63,94,93,39,37,61,48,
	58,113,32,90,44,98,60,51,33,97,62,77,84,80,85,223,
	225,216,187,166,229,189,222,188,141,249,148,200,184,136,248,190,
	199,170,181,204,138,232,218,183,255,234,220,247,213,203,226,193,
	174,172,228,252,217,201,131,230,197,211,145,238,161,179,160,212,
	207,221,254,173,202,146,224,151,140,196,205,130,135,133,143,246,
	192,159,244,239,185,168,215,144,139,165,180,157,147,186,214,176,
	227,231,219,169,175,156,206,198,129,164,150,210,154,177,134,127,
	182,128,158,208,162,132,167,209,149,241,153,251,237,236,171,195,
	243,233,253,240,194,250,191,155,142,137,245,235,163,242,178,152
	};


	/** Communication strings */

	private static final String BEGIN_SSL_CVSNT = "BEGIN SSL AUTH REQUEST"; //$NON-NLS-1$
	private static final String END_SSL_CVSNT = "END SSL AUTH REQUEST"; //$NON-NLS-1$

	private static final String BEGIN_SSL_CVS = "BEGIN SSL REQUEST";//$NON-NLS-1$
	private static final String END_SSL_CVS =   "END SSL REQUEST";//$NON-NLS-1$

	private static final String LOGIN_OK = "I LOVE YOU"; //$NON-NLS-1$
	private static final String LOGIN_FAILED = "I HATE YOU"; //$NON-NLS-1$

	/** End of communication strings */

	private String password;
	private ICVSRepositoryLocation cvsroot;

	private Socket fSocket;

	private InputStream inputStream;
	private OutputStream outputStream;

	// true if using CVSNT sserver, false if using CVS sserver
	private final boolean isCVSNT;
	// the corresponding connection name 
	private String connectionMethodName;
	
	
	// Return the timeout in milliseconds to be used while establishing the SSL connection.
	// Use the CVS communication timeout for this.
	private int getSocketTimeoutWhileHandshaking() {
		// The docs of getTimeout() seem to be wrong. The say milliseconds but it really is numAttempts as used with PollingOutputStream. 
		// With the requested socket timeout value of 1 second this value means seconds, not milliseconds.
		
		// Thus convert seconds to milliseconds!
		return cvsroot.getTimeout() * 1000;
	}
	
	// Timeout value for sockets during regular transfer (milliseconds).
	// Use the CVS communication timeout for this.
	private int getRegularSocketTimeout() {
		// The docs of getTimeout() seem to be wrong. The say milliseconds but it really is numAttempts as used with PollingOutputStream. 
		// With the requested socket timeout value of 1 second this value means seconds, not milliseconds.
		
		// Thus convert seconds to milliseconds!
		return cvsroot.getTimeout() * 1000;
	}

	/**
	 * @see Connection#doClose()
	 */
	public void close() throws IOException {
		try {
			if (inputStream != null)
				inputStream.close();
		} finally {
			inputStream = null;
			try {
				if (outputStream != null)
					outputStream.close();
			} finally {
				outputStream = null;
				try {
					if (fSocket != null)
						fSocket.close();
				} finally {
					fSocket = null;
				}
			}
		}
	}

	/**
	 * @see Connection#doOpen()
	 */
	public void open(IProgressMonitor monitor)
		throws IOException, CVSAuthenticationException {

		monitor.subTask("Authenticating using " +  connectionMethodName); //$NON-NLS-1$
		monitor.worked(1);

		fSocket = createSocket(monitor);
		
		// Start the thread monitoring whether cancelled by user.
		createCancelThread(monitor);
		
		boolean connected = false;
		try {
			createInputOutputStreams(fSocket);
			
			// for CVS sserver we immediately switch to SSL
			if(! isCVSNT) {
				switchToSSL();
			}
			
			requestAuthentication(monitor);
			
			// for CVSNT we must check the response and then switch to SSL
			if(isCVSNT) {
				checkCVSNTAuthRequestResponse();
				switchToSSL();
			}
			
			authenticate(monitor);
			connected = true;
		} finally {
			if (!connected)
				cleanUpAfterFailedConnection();
		}
	}

	private void createInputOutputStreams(Socket s)
		throws IOException {
			
		// We use plain buffered Streams here. In particular we cannot use the PollingInputStream/TimeoutOutputStream 
		// used by PServerConnection, because in contrast to a plain Socket an SSL Socket is _not_ retryable.
		
		if(isDebugStreams()) {
			// debug to System.out
			StreamTracer tracer = new StreamTracer(System.out, new BufferedInputStream(s.getInputStream()), "S>", new BufferedOutputStream(s.getOutputStream()), "C>");
			inputStream = tracer.getTracedInputStream();
			outputStream = tracer.getTracedOutputStream();
		} else {
			inputStream = new BufferedInputStream(s.getInputStream());
			outputStream = new BufferedOutputStream(s.getOutputStream());
		}
	}

	/**
	 * @see Connection#getInputStream()
	 */
	public InputStream getInputStream() {
		return inputStream;
	}
	/**
	 * @see Connection#getOutputStream()
	 */
	public OutputStream getOutputStream() {
		return outputStream;
	}

	/**
	 * Creates a new <code>SSLServerConnection</code> for the given
	 * cvs root.
	 */
	SSLServerConnection(ICVSRepositoryLocation cvsroot, String password, boolean isCVSNT, String connectionMethodName) {
		this.cvsroot = cvsroot;
		this.password = password;
		this.isCVSNT = isCVSNT;
		this.connectionMethodName = connectionMethodName;
	}

	// Request authentication 
	private void requestAuthentication(IProgressMonitor monitor)
		throws IOException, CVSAuthenticationException {

		OutputStream out = getOutputStream();
		StringBuffer request = new StringBuffer();

		// send authentication request
		request.append(getBegin());
		request.append(NEWLINE);
		out.write(request.toString().getBytes());
		out.flush();

	}

	private void checkCVSNTAuthRequestResponse() throws IOException {
		// need to read the last cleartext response from the server while not in SSL
		String sslAuthRequestResponse = readLine(getInputStream());
		
		if (!sslAuthRequestResponse.startsWith("SSERVER")) {
			// remote does not seem to support sserver protocol
			throw new IOException(sslAuthRequestResponse);
		}
	}
	
	// switch to SSL communication
	private void switchToSSL() throws SocketException, IOException {
		// Now switch to SSL, reusing the current plaintext socket
		
		int port = cvsroot.getPort();
		if (port == ICVSRepositoryLocation.USE_DEFAULT_PORT) {
			port = DEFAULT_PORT;
		}
		
		// Save the socket timeout for normal communication		
		final int tempTimeout = fSocket.getSoTimeout();
		
		// Layer SSLSocket over existing plaintext socket
		fSocket =
			SSLPlugin
				.getDefault()
				.getSSLSocketFactory()
				.createSocket(
					fSocket,
					cvsroot.getHost(),
					port,
					true /* close underlying socket when SSL socket is closed*/
		);

		//% The CVS sserver protocol is hardcoded to use v3 mode.
		//% A v2 hello causes the connection to hang, so only allow version 3 protocol.
		if(! isCVSNT) {
			((SSLSocket) fSocket).setEnabledProtocols( new String[] {"SSLv3" } );
		} 
		
		// A socket timeout while handshaking will make client and server lose sync, 
		// so give enough time.
		fSocket.setSoTimeout(getSocketTimeoutWhileHandshaking());
		
		// We will now establish the session and then set the socket timeout value back to the
		// original one. Previously we did this using a HandshakeCompletionListener and the session
		// was established when sending the first data. Unfortunately the Eclipse TimeoutOutputStream 
		// respectively PollingInputStream/PollingOutputStream seemed to interfere with the handshaking,
		// so that in case of a timeout during handshaking an SSLException with message 
		// "Unsupported record version Unknown 20.169" or similar occured instead of a "Connection reset"
		// or "SocketTimeout" error. This is why we now try to establish the SSL session before switching
		// the plaintext socket streams streams to the SSL socket streams.   
		
		// Establish the session
		((SSLSocket)fSocket).startHandshake();
		// Set timeout back to previous value after handshaking.
		fSocket.setSoTimeout(tempTimeout);
		
		
		// switch streams to SSL
		createInputOutputStreams(fSocket);
	}

	// Basically from PServerConnection
	/**
	 * Does the actual authentification.
	 */
	private void authenticate(IProgressMonitor monitor)
		throws IOException, CVSAuthenticationException {

		// socket connection has just been switched to SSL

		String scrambledPassword = scramblePassword(password);
		String user = cvsroot.getUsername();

		OutputStream out = getOutputStream();

		StringBuffer request = new StringBuffer();

		request.append(cvsroot.getRootDirectory());
		request.append(NEWLINE);
		request.append(user);
		request.append(NEWLINE);
		request.append(scrambledPassword);
		request.append(NEWLINE);
		request.append(getEnd());
		request.append(NEWLINE);
		out.write(request.toString().getBytes());
		out.flush();

		String line = readLine(getInputStream());

		// Return if we succeeded
		if (LOGIN_OK.equals(line))
			return;

		// Otherwise, determine the type of error
		if (line.length() == 0)
			throw new IOException(CVSMessages.PServerConnection_noResponse);
		if (LOGIN_FAILED.equals(line))
			throw new CVSAuthenticationException(CVSMessages.PServerConnection_loginRefused, CVSAuthenticationException.RETRY);
		String message = ""; //$NON-NLS-1$
		// Skip any E messages for now
		while (line.charAt(0) == ERROR_CHAR) {
			// message += line.substring(1) + " ";
			line = readLine(getInputStream());
		}
		// Remove leading "error 0"
		if (line.startsWith(ERROR_MESSAGE))
			message += line.substring(ERROR_MESSAGE.length() + 1);
		else
			message += line;
		if (message.indexOf(NO_SUCH_USER) != -1)
			throw new CVSAuthenticationException(NLS.bind(CVSMessages.PServerConnection_invalidUser, (new Object[] {message})), CVSAuthenticationException.RETRY);
		throw new IOException(NLS.bind(CVSMessages.PServerConnection_connectionRefused, (new Object[] { message }))); 
	}

	/*
	 * Called if there are exceptions when connecting.
	 * This method makes sure that all connections are closed.
	 */
	private void cleanUpAfterFailedConnection() throws IOException {
		// same as close()
		close();
	}

	// From PServerConnection
	/**
	 * Creates the actual socket
	 */
	protected Socket createSocket(IProgressMonitor monitor)
		throws IOException {
		// Determine what port to use
		int port = cvsroot.getPort();
		if (port == ICVSRepositoryLocation.USE_DEFAULT_PORT)
			port = DEFAULT_PORT;
		// Make the connection
		Socket result;
		try {
			result = Util.createSocket(cvsroot.getHost(), port, monitor);
		} catch (InterruptedIOException e) {
			// If we get this exception, chances are the host is not responding
			throw new InterruptedIOException(NLS.bind(CVSMessages.PServerConnection_socket, (new Object[] {cvsroot.getHost()}))); 
		}
		result.setSoTimeout(getRegularSocketTimeout());

		return result;
	}

	// From PServerConnection
	private String scramblePassword(String password)
		throws CVSAuthenticationException {
		int length = 0;
		if (password != null) {
			length = password.length();
		}
		char[] out = new char[length];
		for (int i = 0; i < length; i++) {
			char value = password.charAt(i);
			if (value < 0 || value > 255)
				throwInValidCharacter();
			out[i] = SCRAMBLING_TABLE[value];
		}
		return "A" + new String(out); //$NON-NLS-1$
	}

	// From PServerConnection
	private void throwInValidCharacter() throws CVSAuthenticationException {
		throw new CVSAuthenticationException(CVSMessages.PServerConnection_invalidChars, CVSAuthenticationException.RETRY);
	}

	// Copied from org.eclipse.team.internal.ccvs.core.connection.Connection
	private static byte[] append(byte[] buffer, int index, byte b) {
		if (index >= buffer.length) {
			byte[] newBuffer = new byte[index * 2];
			System.arraycopy(buffer, 0, newBuffer, 0, buffer.length);
			buffer = newBuffer;
		}
		buffer[index] = b;
		return buffer;
	}

	// Copied from org.eclipse.team.internal.ccvs.core.connection.Connection
	static String readLine(InputStream in) throws IOException {
		byte[] buffer = new byte[256];
		int index = 0;
		int r;
		while ((r = in.read()) != -1) {
			if (r == NEWLINE)
				break;
			buffer = append(buffer, index++, (byte) r);
		}
		String result = new String(buffer, 0, index);
		if (Policy.DEBUG_CVS_PROTOCOL)
			System.out.println(result);
		return result;
	}
	
	private String getBegin() {
		if(isCVSNT) {
			return BEGIN_SSL_CVSNT;
		} else {
			return BEGIN_SSL_CVS;
		}
	}

	private String getEnd() {
		if(isCVSNT) {
			return END_SSL_CVSNT;
		} else {
			return END_SSL_CVS;
		}
	}
	
	private boolean isClosed() {
		return fSocket == null;
	}
	
	// Creates a thread monitoring if the used pressed cancel. If so, close the socket.
	// This is necessary because otherwise a user might have to wait for the entire
	// CVS communication timeout period until a CVS transaction ends.
	// The PServerConnection uses TimeoutOutputStream/PollingInputStream instead,
	// but these cannot be used for the SSLServerConnection because SSL Sockets are
	// not retryable.
	private void createCancelThread(final IProgressMonitor monitor) {
		Thread cancelThread = new Thread() {
			public void run() {
				// Let this thread run while the connection is not closed.
				while(!isClosed()) {
					// Let this thread sleep for a short period.
					try {
						Thread.sleep(100); // 100 milliseconds
					} catch (InterruptedException e) {
						// Nothing special to do if the thread has been interrupted.
					}
					
					// Check if the monitor has been cancelled.
					if(monitor.isCanceled()) {
						// Give the main thread a chance to react to the cancel request
						// by waiting a few seconds and then checking if the connection has been closed.
						// I consider this to be a more graceful way to than just closing the socket.
						try {
							Thread.sleep(3000); // 3 seconds
						} catch (InterruptedException e) {
							// Nothing special to do if the thread has been interrupted.
						}
						
						// If the connection has already been closed, there is nothing to do.
						// Otherwise we close the connection.
						if(!isClosed()) {
							try {
								// Make sure that fSocket has not been nulled in between.
								// An alternative would be to synchronize #close. 
								Socket tempSocket = fSocket;
								if(tempSocket != null) {
									// Just kill the socket. This will cause a Socket Closed
									// exception in the main thread which will in turn call #close.  
									fSocket.close();
								}
							} catch (IOException e1) {
								// Nothing we could do in this event.
							}
						}
						
						// Terminate the loop and thus the thread.
						break;
					}
				}
			}
		};
		
		cancelThread.start();		
	}
	
	private boolean isDebugStreams() {
		String pluginID = SSLPlugin.PLUGIN_ID;
		boolean debug = "true".equalsIgnoreCase(Platform.getDebugOption(pluginID+"/debug"));
		boolean debugStreams = "true".equalsIgnoreCase(Platform.getDebugOption(pluginID+"/debug/streams"));
		
		return debug && debugStreams;
	}
}